#include "rans_byte.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define SCALE_BITS 10
#define SCALE (1u << SCALE_BITS)

#define HAETAE_MODE 5

#if HAETAE_MODE == 2
#include "samples_h2.h"
#include "samples_z2.h"
#define M_H 13 // Alphabet size for h
static uint32_t f_h[M_H] = {382, 246, 65, 7, 1, 1, 1, 1, 1, 1, 7, 65, 246};
#define M_HB_Z1 13 // Alphabet size for HighBits(z1)
static uint32_t f_hb_z1[M_HB_Z1] = {1, 1, 1, 5, 58, 246, 398, 247, 59, 5, 1, 1, 1};

#elif HAETAE_MODE == 3
#include "samples_h3.h"
#include "samples_z3.h"
#define M_H 17 // Alphabet size for h
static uint32_t f_h[M_H] = {266, 213, 113, 40, 9, 1, 1, 1, 1, 1, 1, 1, 1, 9, 40, 113, 213};
#define M_HB_Z1 17
static uint32_t f_hb_z1[M_HB_Z1] = {1, 1, 1, 1, 8, 37, 112, 216, 269, 216, 112, 38, 8, 1, 1, 1, 1};

#elif HAETAE_MODE == 5
#include "samples_h5.h"
#include "samples_z5.h"
#define M_H 33 // Alphabet size for h
static uint32_t f_h[M_H] = {122, 117, 102, 81, 59, 39, 23, 13, 6, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 6, 13, 24, 39, 59, 81, 102, 118};
#define M_HB_Z1 19
static uint32_t f_hb_z1[M_HB_Z1] = {1, 1, 1, 1, 2, 13, 48, 118, 204, 245, 204, 118, 48, 14, 2, 1, 1, 1, 1};
#endif // HAETAE_MODE

static RansEncSymbol esyms_h[M_H];
static uint16_t symbol_h[SCALE] = {0};
static RansDecSymbol dsyms_h[M_H];

static RansEncSymbol esyms_hb_z1[M_HB_Z1];
static uint16_t symbol_hb_z1[SCALE] = {0};
static RansDecSymbol dsyms_hb_z1[M_HB_Z1];

void print_array(uint32_t* array, size_t length) {
    for(size_t i=0; i< length; i++) {
        printf("%u, ",array[i]);
    }
}

void compute_freqs(uint32_t *freqs, const uint8_t *samples, int sample_size, int m) {
    double tmp;
    int i_max = 0;
    uint32_t freq_max = 0;
    int sum = 0;
    int diff;
    
    // init freqs with 0:
    for(int i=0; i < m; i++)
        freqs[i] = 0;
    // count symbols:
    for(int i=0; i < sample_size; i++) {
        if(samples[i] >= m)
            printf("ERROR: symbol out of range!\n");
        freqs[samples[i]]++;
    }
    // normalize counts:
    for(int i=0; i < m; i++) {
        tmp = freqs[i] / (double) sample_size;
        tmp = tmp * (double) SCALE;
        freqs[i] = (uint32_t) tmp;
        // freq must be at least one:
        if(freqs[i] == 0)
            freqs[i] = 1;
        // track most frequent symbol:
        if(freqs[i] > freq_max) {
            i_max = i;
            freq_max = freqs[i];
        }        
        // compute sum:
        sum += freqs[i];
    }
    // ensure sum == SCALE
    diff = sum - SCALE;
    freqs[i_max] -= diff;

    printf("Freq table with %u symbols, scale %u, diff was %i: {", m, SCALE, diff);
    print_array(freqs, m);
    printf("};\n");
}

void symbol_table(uint16_t *symbol, const uint32_t *freq, size_t alphabet_size) 
{
    int pos = 0;
    for (size_t sym=0; sym < alphabet_size; sym++) {
        for (uint32_t i=0; i < freq[sym]; i++)
            symbol[pos++] = sym;
    }
}

void cum_freq_table(uint32_t *cum_freq, const uint32_t *freq, size_t alphabet_size) {
    cum_freq[0] = 0;
    for (size_t i = 1; i < alphabet_size; i++) {
      cum_freq[i] = cum_freq[i-1] + freq[i-1];
    }
}

void encode_symbols(RansEncSymbol *esyms, const uint32_t *freq, size_t alphabet_size) {
    uint32_t cum_freq[alphabet_size];
    cum_freq_table(cum_freq, freq, alphabet_size);

    for (size_t i=0; i < alphabet_size; i++) 
    {
        RansEncSymbolInit(&esyms[i], cum_freq[i], freq[i], SCALE_BITS);
    }
}

void decode_symbols(RansDecSymbol *dsyms, uint16_t *symbol, const uint32_t *freq, size_t alphabet_size) {
    uint32_t cum_freq[alphabet_size];
    cum_freq_table(cum_freq, freq, alphabet_size);

    symbol_table(symbol, freq, alphabet_size);
    
    for (size_t i=0; i < alphabet_size; i++) 
    {
        RansDecSymbolInit(&dsyms[i], cum_freq[i], freq[i]);
    }
}

void precomputations_rans() {
    compute_freqs(f_h, samples_h, SAMPLE_SIZE_H, M_H);
    encode_symbols(esyms_h, f_h, M_H);
    decode_symbols(dsyms_h, symbol_h, f_h, M_H);

    compute_freqs(f_hb_z1, samples_z, SAMPLE_SIZE_Z, M_HB_Z1);
    encode_symbols(esyms_hb_z1, f_hb_z1, M_HB_Z1);
    decode_symbols(dsyms_hb_z1, symbol_hb_z1, f_hb_z1, M_HB_Z1);
}

void print_esym(RansEncSymbol x){
    printf("{%i, %i, %i, %i, %i}", x.x_max, x.rcp_freq, x.bias, x.cmpl_freq, x.rcp_shift);
}

void print_dsym(RansDecSymbol x){
    printf("{%i, %i}", x.start, x.freq);
}

int main() {
    precomputations_rans();

    printf("\n");

    // encoding h
    printf("static RansEncSymbol esyms_h[M_H] = {");
    for(int i=0; i<M_H; i++) {
        print_esym(esyms_h[i]);
        printf(", ");
    }
    printf("};\n");

    printf("static RansDecSymbol dsyms_h[M_H] = {");
    for(int i=0; i<M_H; i++) {
        print_dsym(dsyms_h[i]);
        printf(", ");
    }
    printf("};\n");

    printf("static uint16_t symbol_h[SCALE] = {");
    for(int i=0; i< (int)SCALE; i++) {
        printf("%i, ", symbol_h[i]);
    }
    printf("};\n");

    // encoding HB(z1)
    printf("static RansEncSymbol esyms_hb_z1[M_HB_Z1] = {");
    for(int i=0; i<M_HB_Z1; i++) {
        print_esym(esyms_hb_z1[i]);
        printf(", ");
    }
    printf("};\n");

    printf("static RansDecSymbol dsyms_hb_z1[M_HB_Z1] = {");
    for(int i=0; i<M_HB_Z1; i++) {
        print_dsym(dsyms_hb_z1[i]);
        printf(", ");
    }
    printf("};\n");

    printf("static uint16_t symbol_hb_z1[SCALE] = {");
    for(int i=0; i< (int)SCALE; i++) {
        printf("%i, ", symbol_hb_z1[i]);
    }
    printf("};\n");

    printf("\n");
}
